// Freeverb port to XMMS.
// v0.01
//
// by Matthew Stewart-Smith
// 21/02/2001

#define _GNU_SOURCE

#include <plugin.h>
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <malloc.h>

static void init(void);
static void cleanup(void);
static void about(void);
static void configure(void);
static int 
mod_samples(gpointer *d, gint length, AFormat afmt, gint srate, gint nch);

// Functions exported from the C++ part of the program
// where the actual freeverb code is located.
extern void do_reverb(float*, float*, float*, float*, long);
extern void set_reverb_params(float, float, float, float, float, float);

static EffectPlugin ep =
{
    NULL,
    NULL,
    "Freeverb 0.01",
    init,
    cleanup,
    about,
    configure,
    mod_samples,
    NULL
};

static const char about_text[] =
"Freeverb reverberation filter.\n"
"Original version (C) 2000 by Jezar.\n"
"http://home.onet.co.uk/~jzracc/freeverb.htm\n\n"
"Ported to XMMS by Matthew Stewart-Smith in Feb. 2001.\n"
"http://www.bigfoot.com/~nimbus186\n"
"nimbus186@bigfoot.com\n\n"
"This plug-in is in the public domain.";

GtkWidget *confwindow;
GtkWidget *aboutwindow;

GtkWidget* hbox[6];
GtkWidget* label[6];
GtkWidget* scale[6];

GtkObject *roomsize_adjustment;
GtkObject *damp_adjustment;
GtkObject *wet_adjustment;
GtkObject *dry_adjustment;
GtkObject *width_adjustment;
GtkObject *atten_adjustment;

float* buffer_in_left;
float* buffer_in_right;
float* buffer_out_left;
float* buffer_out_right;

float attenuation = 0.65;

int conf_stuff_created = 0;

EffectPlugin *get_eplugin_info(void)
{
	return &ep;
}

static void init(void)
{
    buffer_in_left = NULL;
    buffer_in_right = NULL;
    buffer_out_left = NULL;
    buffer_out_right = NULL;
	
    roomsize_adjustment = gtk_adjustment_new(0.50, 0.0, 1.0, 0.05, 0.1, 0);
    damp_adjustment = gtk_adjustment_new(0.50, 0.0, 1.0, 0.05, 0.1, 0);
    wet_adjustment = gtk_adjustment_new(0.25, 0.0, 1.0, 0.05, 0.1, 0);
    dry_adjustment = gtk_adjustment_new(0.95, 0.0, 1.0, 0.05, 0.1, 0);
    width_adjustment = gtk_adjustment_new(1.00, 0.0, 1.0, 0.05, 0.1, 0);
    atten_adjustment = gtk_adjustment_new(0.65, 0.0, 1.0, 0.05, 0.1, 0);
}

static void cleanup(void)
{
    if (buffer_in_left != NULL) free(buffer_in_left);
    if (buffer_in_right != NULL) free(buffer_in_right);
    if (buffer_out_left != NULL) free(buffer_out_left);
    if (buffer_out_right != NULL) free(buffer_out_right);
}

static gint about_destroy_cb(GtkWidget* w, GdkEventAny* e, gpointer data)
{
    gtk_widget_destroy(aboutwindow);
    aboutwindow = NULL;
    return TRUE;
}

static void about_ok_cb(GtkButton* button, gpointer data)
{
    gtk_widget_destroy(aboutwindow);
    aboutwindow = NULL;
}

static void about(void)
{
    GtkWidget *hbox, *text, *button;

    if (aboutwindow != NULL) return;

    aboutwindow = gtk_dialog_new();
    gtk_signal_connect(GTK_OBJECT(aboutwindow), "destroy",
		       GTK_SIGNAL_FUNC(about_destroy_cb), NULL);
    gtk_window_set_title(GTK_WINDOW(aboutwindow), "About Freeverb");

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(aboutwindow)->vbox), hbox,
		       TRUE, TRUE, 5);
    gtk_widget_show(hbox);

    text = gtk_text_new(NULL, NULL);
    gtk_text_set_editable(GTK_TEXT(text), FALSE);
    gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL,
		    about_text, strlen(about_text));
    gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 5);
    gtk_widget_show(text);

    button = gtk_button_new_with_label("OK");
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(aboutwindow)->action_area),
		       button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
		       GTK_SIGNAL_FUNC(about_ok_cb), NULL);

    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(button);
    gtk_widget_show(button);
    gtk_widget_show(aboutwindow);
}


static int mod_samples(gpointer* d, gint length, AFormat afmt, gint srate, gint nch)
{
    int i;
    int no_samples = length >> 2;
    gint16* data = (gint16*) *d;
    gint16* ptr;
    float f;

    // Allocate float sample buffers if this hasn't already been done.
    if (buffer_in_left == NULL)
      {
	  i = sizeof(float) * no_samples;
	  buffer_in_left = (float*) malloc(i);
	  buffer_in_right = (float*) malloc(i);
	  buffer_out_left = (float*) malloc(i);
	  buffer_out_right = (float*) malloc(i);
      }

    /* handle only 16 bit signed stereo native endian */
    if (!(afmt == FMT_S16_NE || (afmt == FMT_S16_LE && G_BYTE_ORDER == G_LITTLE_ENDIAN) || (afmt == FMT_S16_BE && G_BYTE_ORDER == G_BIG_ENDIAN)) || nch != 2)
	return length;

    // Copy data into float input buffer.
    for (i = 0; i < no_samples; i++)
      {
	  buffer_in_left[i]  = ( (float) data[ i<<1     ] ) * attenuation;
	  buffer_in_right[i] = ( (float) data[(i<<1) + 1] ) * attenuation;
      }

    do_reverb(buffer_in_left, buffer_in_right,
	      buffer_out_left, buffer_out_right,
	      no_samples);

    // Extract data from float output buffer.
    for (i = 0; i < no_samples; i++)
      {
	  // Clip to max/min value if overflow/underflow occurs
	  // to reduce audible artifacts resulting from values outside
	  // representable range.

	  f = buffer_out_left[i];
	  if (f > 32767.0)  data[ i<<1     ] = 32767;
	  else if (f < -32768.0) data[ i<<1     ] = -32768;
	  else				   data[ i<<1     ] = f;
	  f = buffer_out_right[i];
	  if (f > 32767.0)  data[(i<<1) + 1] = 32767;
	  else if (f < -32768.0) data[(i<<1) + 1] = -32768;
	  else				   data[(i<<1) + 1] = f;
      }

    return length;
}

static gint conf_destroy_cb(GtkWidget * w, GdkEventAny * e, gpointer data)
{
    int i;
	
    for (i=0;i<6;i++) { gtk_widget_destroy(label[i]); gtk_widget_destroy(hbox[i]); gtk_widget_destroy(scale[i]); }

    gtk_widget_destroy(confwindow);
    confwindow = NULL;
    return TRUE;
}

static void conf_ok_cb(GtkButton * button, gpointer data)
{
    int i;

    attenuation = GTK_ADJUSTMENT(atten_adjustment)->value;
    set_reverb_params(
		      GTK_ADJUSTMENT(roomsize_adjustment)->value,
		      GTK_ADJUSTMENT(damp_adjustment)->value,
		      GTK_ADJUSTMENT(wet_adjustment)->value,
		      GTK_ADJUSTMENT(dry_adjustment)->value,
		      GTK_ADJUSTMENT(width_adjustment)->value,
		      0
		      );

    //	for (i=0;i<6;i++) { gtk_widget_destroy(label[i]); gtk_widget_destroy(hbox[i]); //gtk_widget_destroy(scale[i]); }
	
    //	gtk_widget_destroy(confwindow);
    //	confwindow = NULL;
}

static void conf_cancel_cb(GtkButton * button, gpointer data)
{
    int i;
	
    for (i=0;i<6;i++) { gtk_widget_destroy(label[i]); gtk_widget_destroy(hbox[i]); gtk_widget_destroy(scale[i]); }

    gtk_widget_destroy(confwindow);
    confwindow = NULL;
}

static void configure(void)
{
    GtkWidget* button;
    int i;

    if (confwindow != NULL)
	return;

    confwindow = gtk_dialog_new();
    gtk_signal_connect(GTK_OBJECT(confwindow), "destroy",
		       GTK_SIGNAL_FUNC(conf_destroy_cb), NULL);
    gtk_window_set_title(GTK_WINDOW(confwindow), "Configure Freeverb");

    label[0] = gtk_label_new("Room size");
    label[1] = gtk_label_new("Damping");
    label[2] = gtk_label_new("Wet level");
    label[3] = gtk_label_new("Dry level");
    label[4] = gtk_label_new("Width");
    label[5] = gtk_label_new("Attenuation");

    for (i=0;i<6;i++) hbox[i] = gtk_hbox_new(FALSE, 10);	

    scale[0] = gtk_hscale_new(GTK_ADJUSTMENT(roomsize_adjustment));
    scale[1] = gtk_hscale_new(GTK_ADJUSTMENT(damp_adjustment));
    scale[2] = gtk_hscale_new(GTK_ADJUSTMENT(wet_adjustment));
    scale[3] = gtk_hscale_new(GTK_ADJUSTMENT(dry_adjustment));
    scale[4] = gtk_hscale_new(GTK_ADJUSTMENT(width_adjustment));
    scale[5] = gtk_hscale_new(GTK_ADJUSTMENT(atten_adjustment));

    for (i=0;i<6;i++)
      {
	  gtk_scale_set_digits((GtkScale*)(scale[i]), 2);
	  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(confwindow)->vbox), hbox[i],
			     TRUE, TRUE, 10);
	  gtk_widget_show(hbox[i]);
	  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(confwindow)->vbox), label[i],
			     TRUE, TRUE, 0);
	  gtk_widget_show(label[i]);
	  gtk_box_pack_start(GTK_BOX(hbox[i]), scale[i], TRUE, TRUE, 10);
	  gtk_widget_show(scale[i]);
      }

    button = gtk_button_new_with_label("Ok");
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(confwindow)->action_area),
		       button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
		       GTK_SIGNAL_FUNC(conf_ok_cb), NULL);
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default(button);
    gtk_widget_show(button);

    gtk_widget_show(confwindow);
}
